from pathlib import Path
from types import ModuleType


def import_module_from(path: Path) -> ModuleType:
    """
    Import a Python module into runtime.

    :param path: a path or string that represents a Python module.

    >>> import_module_from(Path(""))
    Traceback (most recent call last):
    ...
    ModuleNotFoundError: Module 'None' not found.
    """
    import importlib.util

    if not path or not path.exists():
        raise ModuleNotFoundError(f"There is no module at '{path.resolve()}'.")

    name = path.name.replace(".py", "")  # get the module name (filename without extension)
    module_spec = importlib.util.spec_from_file_location(name, path)  # get the module specification

    if module_spec:
        module_instance = importlib.util.module_from_spec(module_spec)

        if module_instance:
            module_spec.loader.exec_module(module_instance)
        else:
            raise ModuleNotFoundError(f"Module spec '{module_spec}' not found.")
    else:
        raise ModuleNotFoundError(f"Module '{module_spec}' not found.")

    return module_instance


def import_package_from(path: Path) -> list[ModuleType]:
    """
    Manually load all Python modules inside a package.
    """
    return [import_module_from(Path(module_path)) for module_path in path.glob("*.py")]


def import_module(name: str) -> ModuleType:
    """
    Import/Load a python module.

    :param name: module name, e.g. `"package.module"`
    """
    import importlib
    return importlib.import_module(name)


def import_modules(modules: list[str]) -> list[ModuleType]:
    """
    Import/Load a list of python modules.

    :param modules: a list of module names, e.g. `["package.module"]`
    """
    return [import_module(module_name) for module_name in modules]



from datetime import datetime
from decimal import Decimal
from typing import Optional

from pydantic import BaseModel, computed_field, Field


class _WiseQuote(BaseModel):
    rate: Decimal
    fee: Decimal
    dateCollected: datetime
    sourceCountry: Optional[str]
    targetCountry: Optional[str]
    markup: Decimal
    receivedAmount: Optional[Decimal]
    sendAmount: Optional[Decimal]
    isConsideredMidMarketRate: bool

    @computed_field
    @property
    def amount(self) -> Decimal:
        if self.sendAmount:
            return self.sendAmount
        elif self.receivedAmount:
            return self.receivedAmount
        else:
            return Decimal(0)


class _WiseQuotationProvider(BaseModel):
    id: int
    alias: str
    name: str
    logos: dict
    type: str
    partner: bool
    quotes: list[_WiseQuote]


class _WiseQuotationResponse(BaseModel):
    sourceCurrency: str
    targetCurrency: str
    sourceCountry: Optional[str] = None
    targetCountry: Optional[str] = None
    providerCountry: Optional[str] = None
    providerTypes: list[str]
    amount: Decimal
    amountType: str
    providers: list[_WiseQuotationProvider]

    @computed_field
    @property
    def quotation(self) -> _WiseQuote:
        provider: Optional[_WiseQuotationProvider] = None
        quote: Optional[_WiseQuote] = None

        if self.providers:
            for provider in self.providers:
                if provider.alias == "wise":
                    break

        if provider and len(provider.quotes) >= 1:
            quote = provider.quotes[0]

        return quote


class CurrencyQuote(BaseModel):
    source_currency: str
    target_currency: str
    date: datetime
    rate: Decimal


class TransferQuote(BaseModel):
    provider: str = "wise"
    quote: CurrencyQuote
    sourceCountry: Optional[str]
    targetCountry: Optional[str]
    date: datetime = Field(validation_alias="dateCollected")
    amount: Decimal
    fee: Decimal

    @computed_field
    @property
    def no_fee_amount(self) -> Decimal:
        return self.amount - self.fee



from typing import Optional

import httpx

from commons.media.images import Image


def avatar(key: str, size: int) -> Optional[Image]:
    response = httpx.get(f'https://www.gravatar.com/avatar/{key}?s={size}')

    if response and response.content:
        return Image(data=response.content).convert()
    else:
        raise ConnectionError(
            f"An error has occurred while fetching Wise API: {response.status_code} - {response.content}")





def build(params: dict) -> str:
    """
    Build an HTTP Query parameters string from dictionary.
    :param params: dictionary with all parameters key-value pairs
    :return: a query string
    """
    return f"{'&'.join([f'{str(key)}={str(val)}' for key, val in params.items() if val])}"



from asyncio import Lock
from datetime import datetime
from typing import Optional

from pydantic import BaseModel, SecretStr, Field, field_serializer, PrivateAttr
from sqlalchemy.engine.base import Engine
from sqlmodel import Session, create_engine


class SqlDatetimeAdapter:
    """
    Adapter inspired by `web.py` [1] after deprecation of default datetime adapters for SQLite on Python v3.12 [2].

    [1] [web.py PR !786]:
    https://github.com/webpy/webpy/pull/786
    [2] [Deprecate default (built-in) sqlite3 converters and adapters  - Python.org]:
    https://discuss.python.org/t/deprecate-default-built-in-sqlite3-converters-and-adapters/15781/16
    """

    @staticmethod
    def to_timestamp(date_time: datetime) -> str:
        """
        Convert a Python datetime.datetime into a timezone-naive ISO 8601 date string.

        >>> SqlDatetimeAdapter.to_timestamp(datetime(2024, 9, 10, 10, 10, 0, 12))
        "2024-09-10T10:10:0.000012"
        """
        return date_time.isoformat()

    @staticmethod
    def to_datetime(timestamp: bytes) -> datetime:
        """
        Convert an ISO 8601 formatted bytestring to a datetime.datetime object.

        >>> SqlDatetimeAdapter.to_datetime(b"2024-09-10T10:10:0.000012")
        datetime.datetime(2024, 9, 10, 10, 10, 0, 12)
        """
        return datetime.strptime(timestamp.decode("utf-8"), "%Y-%m-%dT%H:%M:%S.%f")


class DatabaseAdapter(BaseModel):
    """
    Thread-Safe SQLAlchemy-like URL connector adapter.

    This class provides a way to connect to a database using a URL.
    It supports various database drivers, including SQLite, PostgreSQL, and MySQL.
    """
    scheme: str
    username: Optional[str] = None
    password: Optional[SecretStr] = None
    host: str = ""
    port: Optional[int] = None
    database: str = Field(serialization_alias="path", default="sqlite.db")
    query: str = ""

    _engine: Optional[Engine] = PrivateAttr(default=None)

    @field_serializer("password")
    def dump_secret(self, v):
        if v:
            return v.get_secret_value()

    # to use a threading.Lock object with Pydantic, allow the class to have arbitrary types and
    # use Field to initialize the lock attribute
    # Source: https://stackoverflow.com/a/77151583
    # Docs:   https://docs.pydantic.dev/1.10/usage/model_config/
    # Licensed under CC BY-SA
    class Config:
        arbitrary_types_allowed = True

    lock: Lock = Field(exclude=True, default=Lock())
    # --- end of code snippet ---

    def uri(self) -> str:
        """
        Provide the URI of a SQL database. It does dump the database password as plain-text.
        """
        from pydantic_core import MultiHostUrl

        return MultiHostUrl.build(
            **self.model_dump(by_alias=True)
        ).unicode_string()

    async def _make_engine(self) -> Engine:
        engine: Engine
        async with self.lock:
            if self.scheme.startswith("sqlite"):
                import sqlite3

                sqlite3.register_adapter(datetime, SqlDatetimeAdapter.to_timestamp)
                sqlite3.register_converter("timestamp", SqlDatetimeAdapter.to_datetime)

                engine = create_engine(self.uri())

        return engine

    def engine(self) -> Engine:
        if not self._engine:
            # --- Code from: https://stackoverflow.com/a/75341431; License: CC BY-SA 4.0
            import asyncio
            from concurrent.futures import ThreadPoolExecutor

            try:
                asyncio.get_running_loop()  # Triggers RuntimeError if no running event loop
                # Create a separate thread so we can block before returning
                with ThreadPoolExecutor(1) as pool:
                    self._engine = pool.submit(lambda: asyncio.run(self._make_engine())).result()
            except RuntimeError:
                self._engine = asyncio.run(self._make_engine())
            # --- end of licensed code

        return self._engine

    def session(self) -> Session:
        return Session(self.engine())



import unittest


class TestMIMETypes(unittest.TestCase):
    def test_lookup(self):
        from commons.media import mimetypes

        assert mimetypes.lookup(".png") == mimetypes.IMAGE_PNG
        assert mimetypes.lookup("jpg") == mimetypes.IMAGE_JPEG
        assert mimetypes.lookup(".PDF") == mimetypes.APPLICATION_PDF
        assert mimetypes.lookup("json") == mimetypes.APPLICATION_JSON
        assert mimetypes.lookup("image.webp") == mimetypes.IMAGE_WEBP
        assert mimetypes.lookup("http://example.com/image.png") == mimetypes.IMAGE_PNG



from typing import Optional

from pydantic import BaseModel, field_validator

from commons.media import Image, mimetypes

"""
Module adapted from PyPix package by Artemis Resende.

Source: https://github.com/cleitonleonel/pypix/
License: MIT License

Copyright (c) 2022 Cleiton Leonel Creton

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# noinspection PyNestedDecorators
class Pix(BaseModel):
    """
    Model for Brazilian Payment System PIX.
    """
    single_transaction: bool = False
    key: str
    receiver_name: str
    receiver_city: str = ""
    amount: float = 0.0
    receiver_zipcode: Optional[str] = None
    identification: Optional[str] = None
    description: Optional[str] = None
    default_url_pix: Optional[str] = None

    @classmethod
    def _cpf_key(cls, source: str) -> Optional[str]:
        is_valid: bool = True
        numbers = [int(char) for char in source if char.isdigit()]

        if (len(numbers) != 11) or (numbers == numbers[::-1]):
            is_valid = False

        try:
            for i in range(9, 11):
                value = sum((numbers[num] * ((i + 1) - num) for num in range(0, i)))
                digit = ((value * 10) % 11) % 10
                if digit != numbers[i]:
                    is_valid = False
        except (IndexError, Exception):
            is_valid = False

        return source if is_valid else None

    @classmethod
    def _phone_key(cls, source: Optional[str]) -> Optional[str]:
        import re

        rule = re.compile(r'^\+?[1-9]\d{1,14}$')

        if rule.match(source):
            if source.startswith("+55"):
                return source
            else:
                return f"+55{source}"
        else:
            return None

    @field_validator("key", mode="before")
    @classmethod
    def key_validator(cls, value: str) -> str:
        key = cls._cpf_key(value)  # check if it is a CPF

        if not key:
            key = cls._phone_key(value)  # check if it is a brazilian phone number

        if not key:
            # todo: add validation for random and e-mail keys
            key = value

        return key

    @field_validator("default_url_pix", mode="before")
    @classmethod
    def validate_default_url_pix(cls, value: Optional[str]) -> Optional[str]:
        return value.replace('https://', '') if value else None

    @field_validator("receiver_name", mode="before")
    @classmethod
    def validate_receiver_name(cls, value: Optional[str]) -> Optional[str]:
        if len(value) > 25:
            raise ValueError('The maximum number of characters for the receiver name is 25.')
        return value

    @field_validator("receiver_city", mode="before")
    @classmethod
    def validate_receiver_city(cls, value: Optional[str]) -> Optional[str]:
        if len(value) > 15:
            raise ValueError('The maximum number of characters for the receiver city is 15.')
        return value

    @field_validator("amount", mode="before")
    @classmethod
    def validate_amount(cls, value: float) -> float:
        if len(str("{0:.2f}".format(value))) > 13:
            raise ValueError('The maximum number of characters for the amount value is 13.')
        return float("{0:.2f}".format(value))

    def brcode(self):
        """
        Generate a BR code [1].

        [1]: The BR Code is the name of the QR Code standard, for the purpose of initiating payments, adopted in Brazil,
         in accordance with Circular No. 3,682, dated November 4, 2013.

        Source: https://www.bcb.gov.br/content/estabilidadefinanceira/spb_docs/ManualBRCode.pdf

        :return: BR Code data as string.
        """
        def encode_value(identifier: str, value: str) -> str:
            return f"{identifier}{str(len(value)).zfill(2)}{value}"

        def format_text(value: str) -> str:
            from unicodedata import normalize
            import re
            if value:
                return re.sub(
                    r'[^A-Za-z0-9$@%*+\-./:_ ]', '',
                    normalize('NFD', value).encode('ascii', 'ignore').decode('ascii')
                )
            else:
                return ""

        def encode_info():
            base_pix = encode_value('00', 'br.gov.bcb.pix')
            info_string = ''

            if self.key:
                info_string += encode_value('01', self.key)
            elif self.default_url_pix:
                info_string += encode_value('25', self.default_url_pix)
            else:
                raise ValueError('You must enter a URL or a Pix key to generate a valid BR Code.')

            if self.description:
                info_string += encode_value('02', format_text(self.description))

            return encode_value('26', f'{base_pix}{info_string}')

        def encode_additional_data():
            if self.identification:
                return encode_value('62', encode_value('05', format_text(self.identification)))
            return encode_value('62', encode_value('05', '***'))

        def compute_crc(hex_code: str) -> str:
            from binascii import crc_hqx
            msg = bytes(hex_code, "utf-8")
            crc = crc_hqx(msg, 0xffff)
            return "{:04X}".format(crc & 0xffff)

        result_string = (
            f"{encode_value("00", "01")}"
            f"{encode_value("01", "12" if self.single_transaction else "11")}"
            f"{encode_info()}"
            f"{encode_value("52", "0000")}"
            f"{encode_value("53", "986")}"
            f"{encode_value("54", str(self.amount))}"
            f"{encode_value("58", "BR")}"
            f"{encode_value("59", format_text(self.receiver_name))}"
            f"{encode_value("60", format_text(self.receiver_city))}"
            f"{encode_value("61", format_text(self.receiver_zipcode))}"
            f"{encode_additional_data()}"
            f"6304"
        )
        return result_string + compute_crc(result_string)

    def qrcode(self, color: str = "black", back_color: str = "white",
               amount: Optional[float] = None,
               description: Optional[str] = None) -> Image:
        from commons.currencies.payments import qrcode
        return qrcode.generate(currency="BRL", wallet=self.key, receiver=self.receiver_name, amount=amount,
                        description=description, color=color, back_color=back_color)



import json
from pathlib import Path

import pycountry

_CRYPTO_CURRENCIES: list[dict] = [
    {"alpha_3": "BTC", "name": "Bitcoin"},
    {"alpha_3": "XMR", "name": "Monero"},
]
_iso4217_codes = json.loads((Path(pycountry.DATABASE_DIR) / "iso4217.json").read_bytes())["4217"]


for _v in _CRYPTO_CURRENCIES + _iso4217_codes:
    globals()[_v["alpha_3"]] = _v["name"]


def is_crypto(currency_code: str) -> bool:
    return bool(currency_code in [currency["alpha_3"] for currency in _CRYPTO_CURRENCIES])


# noinspection PyUnresolvedReferences
def lookup(currency: str) -> str:
    """
    Lookup for a currency code. It supports Bitcoin.
    :param currency: a currency representation
    :return: an Alpha-3 ISO-4217 upper-case code
    """
    try:
        if currency:
            currency = currency.upper()
        else:
            raise LookupError

        return currency if globals().get(currency) else None
    except LookupError:
        raise ValueError(f"'{currency}' is neither a crypto or ISO-4217 (Alpha-3) currency code.")



from typing import Optional

from commons import currencies
from commons.media import Image


def generate(currency: str, wallet: str,
             receiver: Optional[str] = None,
             amount: Optional[float] = None,
             description: Optional[str] = None,
             color: str = "black", back_color: str = "white") -> Optional[Image]:
    def qrcode(data: str, box_size: int = 8):
        from qrcode.main import QRCode
        from qrcode.image.pure import PyPNGImage
        from qrcode.image.pil import PilImage
        from commons.media import Image, mimetypes
        from io import BytesIO

        code: QRCode = QRCode(box_size=box_size)
        buffer: BytesIO = BytesIO()
        image: PilImage | PyPNGImage

        code.add_data(data)
        code.make(fit=True)

        image = code.make_image(fill_color=color, back_color=back_color).convert("RGB")
        image.save(buffer, 'png')

        return Image(data=buffer.getvalue(), media_type=mimetypes.IMAGE_PNG)

    # ---
    currency = currencies.lookup(currency)

    if wallet:
        match currency:
            case "BTC":
                from commons.network.http import query

                params = {
                    "amount": f"{amount:.8f}" if amount else None,
                    "label": receiver if receiver else None,
                    "message": description if description else None
                }

                q = query.build(params)
                if q:
                    return qrcode(f"bitcoin:{wallet}?{q}")
                else:
                    return qrcode(f"bitcoin:{wallet}")
            case "XMR":
                from commons.network.http import query

                params = {
                    "tx_amount": f"{amount:.12f}" if amount else None,
                    "recipient_name": receiver if receiver else None,
                    "tx_description": description if description else None
                }

                q = query.build(params)
                if q:
                    return qrcode(f"monero:{wallet}?{q}")
                else:
                    return qrcode(f"monero:{wallet}")
            case "BRL":
                if receiver and wallet:
                    from commons.currencies.payments import Pix

                    params = {
                        "receiver_name": receiver,
                        "key": wallet,
                        "amount": amount if amount else 0,
                        "description": description if description else ""
                    }

                    return qrcode(data=Pix(**params).brcode(), box_size=5)
                else:
                    raise ValueError("BRL QR code via Pix requires a receiver and a wallet.")
            case _:
                return None
    else:
        raise ValueError("No wallet was provided.")



import unittest

import commons.currencies.payments.qrcode
from commons import currencies


class TestCurrencies(unittest.TestCase):
    def test_currency_client(self):
        assert currencies.lookup("brl") == "BRL"
        assert currencies.lookup("eur") == "EUR"
        assert currencies.lookup("btc") == "BTC"
        assert currencies.lookup("xmr") == "XMR"
        self.assertRaises(ValueError, currencies.lookup, currency="")

    def test_qrcode_gen(self):
        assert commons.currencies.payments.qrcode.generate(currency="brl", wallet="pix@email.com", receiver="Maria Joana")
        assert commons.currencies.payments.qrcode.generate(currency="brl", wallet="pix@email.com", receiver="Maria Joana", amount=100, description="pagamento")
        assert commons.currencies.payments.qrcode.generate(currency="btc", wallet="samplekey")
        assert commons.currencies.payments.qrcode.generate(currency="btc", wallet="samplekey", receiver="Maria Joana", amount=100, description="pagamento")
        assert commons.currencies.payments.qrcode.generate(currency="xmr", wallet="samplekey")
        assert commons.currencies.payments.qrcode.generate(currency="xmr", wallet="samplekey", receiver="Maria Joana", amount=100, description="pagamento")
        self.assertRaises(ValueError, commons.currencies.payments.qrcode.generate, currency="BRL", wallet="pix@email.com")
        self.assertRaises(ValueError, commons.currencies.payments.qrcode.generate, currency="BRL", wallet="")



import tempfile
import unittest
from pathlib import Path
from time import sleep

from sqlmodel import SQLModel, select

from commons.database import DatabaseAdapter, DatabaseMigrationExecutor

RESOURCES_FOLDER: Path = Path(__file__).parent / "resources"


class TestDatabase(unittest.TestCase):
    @staticmethod
    def _get_db() -> DatabaseAdapter:
        return DatabaseAdapter(
            scheme="sqlite",
            database=tempfile.mktemp(dir=tempfile.gettempdir())
        )

    def test_database_creation(self):
        database: DatabaseAdapter = self._get_db()
        session = database.session()

        assert database
        assert database.engine()
        assert session
        session.close()

    def test_migration_files(self):
        from tests import TestUser

        db = self._get_db()
        SQLModel.metadata.create_all(db.engine())

        with db.session() as session:
            data = DatabaseMigrationExecutor(path=RESOURCES_FOLDER / "migrations", session=session).run()

            assert data
            assert session.exec(select(TestUser).where(TestUser.username == "john.doe")).first()
            assert session.exec(select(TestUser).where(TestUser.username == "joana.doe")).first()

            session.close()

    def test_cache(self):
        from commons.database.cache import Cache

        db = self._get_db()
        SQLModel.metadata.create_all(db.engine())

        with Cache(db) as cache:
            entry = cache.set("key", b"value", max_age=1)
            assert entry
            assert not entry.expired
            sleep(1)
            assert cache.get("key").expired
            cache.invalidate("key")
            assert not cache.get("key")

    def test_async_db(self):
        import asyncio
        async def _inner_func():
            return self._get_db().engine()

        asyncio.run(_inner_func())




from pathlib import Path
from typing import Optional

import httpx
from pydantic import computed_field, ConfigDict
from sqlmodel import SQLModel, Field


# noinspection PyNestedDecorators
class Resource(SQLModel):
    """
    Represents and loads a resource.
    Be aware that might be unsafe to read remote content that you might not trust.
    """
    model_config = ConfigDict(
        arbitrary_types_allowed=True
    )

    location: str = Field(primary_key=True, nullable=False, default="memory://")
    data: Optional[bytes] = Field(nullable=True, default=None)

    @computed_field
    @property
    def url(self) -> httpx.URL:
        return httpx.URL(self.location)

    @computed_field
    @property
    def path(self) -> Path:
        return Path(self.location)

    def is_remote(self) -> bool:
        """
        Check whether a location is remote or local
        """
        return self.url.is_absolute_url

    def is_local(self) -> bool:
        """
        Check whether a location is local or remote
        """
        return self.url.is_relative_url

    def exists(self) -> bool:
        """Check if the resource exists"""
        if self.is_local():
            return self.path.exists()
        elif self.is_remote():
            return bool(self.read())
        else:
            return bool(self._data)

    def read(self) -> Optional[bytes]:
        """Read a resource, either locally or remotely (via HTTP)."""
        if not self.data:
            try:
                if self.is_remote():
                    response = httpx.get(self.url, headers={"Accept-Encoding": "utf-8"})

                    if response and response.status_code == 200:
                        self.data = response.read()
                elif self.exists():
                    self.data = self.path.read_bytes()
            except FileNotFoundError:
                ...

        return self.data

    def scheme(self) -> str:
        return self.url.scheme

    def filename(self) -> Optional[str]:
        return self.path.name if self.path else None



import unittest

from commons.media import Image


class TestHTTPClients(unittest.TestCase):
    def test_currency_client(self):
        from commons import currencies

        assert currencies.get_quote(from_currency="BRL", to_currency="EUR")
        assert currencies.get_quote(from_currency="BTC", to_currency="BRL")
        assert currencies.get_quote(from_currency="XMR", to_currency="EUR")

    def test_gravatar_client(self):
        from commons.media import gravatar

        image: Image = gravatar.avatar("e5f43fe12e80783bd2666c529fbf33d0", size=120)
        assert image and image.read()

    def test_giphy_client(self):
        from commons.media import giphy

        image: Image = giphy.gif("l4HogOSqU3uupmvmg")
        assert image and image.read()



from typing import Optional

import httpx

from commons.media import Image


def gif(key: str) -> Optional[Image]:
    response = httpx.get(f"https://media.giphy.com/media/{key}/giphy.gif")

    if response and response.content:
        return Image(data=response.content, media_type=response.headers['Content-Type'])
    else:
        raise ConnectionError(
            f"An error has occurred while fetching Wise API: {response.status_code} - {response.content}")



import logging
from pathlib import Path

from command_runner import command_runner
from pydantic import BaseModel
from commons.network import IPV4_LOOPBACK


class CertificateFiles(BaseModel):
    domain: str = IPV4_LOOPBACK
    cert: Path
    key: Path


def get_cert(save_to: Path, domains: list[str] | None = None) -> dict[str, CertificateFiles] | CertificateFiles:
    """
    Create an X.509 certificate with key. If no domain is provided, a local certificate is created for 0.0.0.0 (loopback)

    :param save_to: file destination path
    :param domains: list of domains
    :return: location of the certificates
    """

    domains = domains if domains else [IPV4_LOOPBACK]
    generated_certs: dict[str, CertificateFiles] = {}
    # TODO: implement certbot

    if not save_to.exists():
        import os
        os.mkdir(save_to)

    for domain in domains:
        file: CertificateFiles = CertificateFiles(
            cert=Path(f"{save_to / domain}.pem"),
            key=Path(f"{save_to / domain}-key.pem")
        )

        cmd = f"mkcert -cert-file {file.cert.resolve()} -key-file {file.key.resolve()} {domain}"

        logging.getLogger(__file__).debug(f"Full command: '{cmd}'")

        exit_code, output = command_runner(cmd, silent=True)

        if exit_code == 253:
            raise RuntimeError("'mkcert' not found.")

        generated_certs[domain] = file

    return generated_certs if len(generated_certs) > 1 else generated_certs[domains[0]]



"""
Implements a client that fetches currency data from Wise and Coinbase.
Documentation: https://docs.wise.com/api-docs/api-reference/comparison
"""

import json
from datetime import datetime
from decimal import Decimal
from logging import getLogger, Logger
from typing import Optional

import httpx

from commons.currencies._models import _WiseQuotationResponse, CurrencyQuote, TransferQuote
from commons.currencies import codes as currencies
from commons.media import Image
from commons.network.http import query

logger: Logger = getLogger(__file__)


# noinspection PyUnresolvedReferences
def get_transfer_quote(from_currency: str, to_currency: str,
                       send_amount: Optional[Decimal] = None,
                       recipient_gets_amount: Optional[Decimal] = None,
                       source_country: Optional[str] = None, target_country: Optional[str] = None,
                       decimal_precision: int = 2) -> TransferQuote:
    """
    Get a transfer quotation between two currencies. Either `send_amount` or `recipient_gets_amount` should be specified.
    It does not support cryptocurrencies.

    :param from_currency: ISO-4217 currency as string
    :param to_currency: ISO-4217 currency as string
    :param send_amount: [Optional] Amount to send
    :param recipient_gets_amount: [Optional] Amount to receive
    :param source_country: [Optional] ISO-3166-1 Alpha-2 country code as string
    :param target_country: [Optional] ISO-3166-1 Alpha-2 country code as string
    :param decimal_precision: [Optional] Precision used on amount conversion. Default is 2.
    """
    # --- Build parameters
    params: dict = {
        "sourceCurrency": currencies.lookup(from_currency),
        "targetCurrency": currencies.lookup(to_currency)
    }

    if send_amount:
        params["sendAmount"] = f"{send_amount:.{decimal_precision}f}"
    elif recipient_gets_amount:
        params["recipientGetsAmount"] = f"{recipient_gets_amount:.{decimal_precision}f}"
    else:
        raise ValueError("Either `send_amount` or `recipient_gets_amount` must be specified.")

    if source_country:
        params["sourceCountry"] = countries.lookup(source_country).alpha_2
    if target_country:
        params["targetCountry"] = countries.lookup(target_country).alpha_2

    # --- Request data
    response = httpx.get(f"https://api.wise.com/v4/comparisons/?{query.build(params)}")
    if response and response.status_code == 200:
        # build response
        wise_response: _WiseQuotationResponse = _WiseQuotationResponse(**json.loads(response.content.decode()))
        quote: CurrencyQuote = CurrencyQuote(
            **{"source_currency": wise_response.sourceCurrency,
               "target_currency": wise_response.targetCurrency,
               "date": wise_response.quotation.dateCollected,
               "rate": wise_response.quotation.rate}
        )
        transfer_quote: TransferQuote = TransferQuote(**({"quote": quote} | wise_response.quotation.model_dump()))

        return transfer_quote
    else:
        raise ConnectionError(f"An error has occurred while fetching Wise API: {response.status_code} - {response.content}")


def _get_iso4217_quote(from_currency: str, to_currency: str) -> CurrencyQuote:
    return get_transfer_quote(from_currency, to_currency, recipient_gets_amount=Decimal(1)).quote


def _get_crypto_quote(from_currency: str, to_currency: str) -> CurrencyQuote:
    # docs: https://developers.coindesk.com/documentation/legacy/Price/SingleSymbolPriceEndpoint/
    response = httpx.get(f"https://min-api.cryptocompare.com/data/price?fsym={from_currency}&tsyms={to_currency}")
    if response and response.status_code == 200:
        return CurrencyQuote(
            **{"source_currency": from_currency,
               "target_currency": to_currency,
               "date": datetime.now(),
               "rate": json.loads(response.content.decode())[to_currency]
               }
        )
    else:
        raise ConnectionError(
            f"An error has occurred while fetching CryptoCompare API: {response.status_code} - {response.content}")


def get_quote(from_currency: str, to_currency: str) -> CurrencyQuote:
    """
    Get currency rate between two currencies. It supports BTC and XMR among other currencies.

    :param from_currency: ISO-4217 currency as string or cryptocurrency representation
    :param to_currency: ISO-4217 currency as string or cryptocurrency representation
    """
    from_currency = currencies.lookup(from_currency)
    to_currency = currencies.lookup(to_currency)

    if currencies.is_crypto(from_currency) or currencies.is_crypto(to_currency):
        return _get_crypto_quote(from_currency, to_currency)
    else:
        return _get_iso4217_quote(from_currency, to_currency)



# test_log_config.py
import re
import tempfile
import unittest
import logging
from pathlib import Path
import os
from commons.logging import config, get_logger


class TestLogConfig(unittest.TestCase):
    message_pattern = r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}.\d{3} - \w+ \[(DEBUG|INFO|WARNING|ERROR|CRITICAL)\]: .+$"

    @classmethod
    def setUpClass(cls):
        cls.log_dir = Path(tempfile.mkdtemp())
        cls.log_file = cls.log_dir / "root.log"
        cls.logger = config(
            level=logging.DEBUG,
            directory=cls.log_dir,
            max_file_bytes=1024,
            backup_count=3,
            stream=None  # disable stdout
        )

    @classmethod
    def tearDownClass(cls):
        os.remove(cls.log_file)
        os.rmdir(cls.log_dir)

    def test_logger_creation(self):
        logger = get_logger("TestLogger")
        self.assertIsInstance(logger, logging.Logger)
        self.assertEqual(logger.name, "TestLogger")

    def test_log_file_creation(self):
        self.logger.info("Test log message")
        self.assertTrue(self.log_file.exists())

    def test_log_level(self):
        logger = get_logger()
        logger.debug("Debug message")
        logger.info("Info message")
        logger.warning("Warning message")
        logger.error("Error message")

        self.assertEqual(logger, self.logger)  # check if it is the root logger
        self.assertEqual(logger.level, logging.DEBUG)
        for log in self.log_file.read_text().splitlines():
            self.assertTrue(re.match(self.message_pattern, log))

    def test_global_logging_config(self):
        from logging.handlers import RotatingFileHandler

        self.assertEqual(self.logger.level, logging.DEBUG)
        self.assertTrue(
            any(isinstance(handler, RotatingFileHandler) for handler in self.logger.handlers))

    def test_custom_logger_config(self):
        custom_logger = get_logger(
            name="CustomLogger",
            level=logging.WARNING,
        )
        self.assertEqual(custom_logger.level, logging.WARNING)
        self.assertEqual(custom_logger.name, "CustomLogger")



from io import BytesIO
from pathlib import Path
from typing import Optional

from httpx import URL

from commons.media import mimetypes
from commons.models import Resource


class Processor:
    def __init__(self, data: bytes):
        from PIL import Image

        self.data = data
        self.image = Image.open(BytesIO(data))  # Create an Image based on an in-memory image bytestream
        self.buffer = BytesIO()

    def compress(self):
        self.image.save(self.buffer, self.image.format, optimize=True, quality=9)

        return self

    def convert(self, to_type: str = mimetypes.IMAGE_WEBP):
        _type = to_type

        match _type:
            case mimetypes.IMAGE_WEBP:
                self.image.save(self.buffer, "WEBP")
            case mimetypes.APPLICATION_PDF:
                import img2pdf
                self.buffer.write(img2pdf.convert(self.data))
            case _:
                raise ValueError(f'\'{to_type}')

        return self

    def get(self):
        self.buffer.seek(0)  # Move the bytestream pointer to the beginning

        return self.buffer.read()


# noinspection PyUnresolvedReferences
class Image(Resource):
    alt: Optional[str] = None

    def __init__(self, location: Optional[str | Path | URL] = None,
                 alt: Optional[str] = None,
                 width: Optional[int] = None,
                 height: Optional[int] = None,
                 data: Optional[bytes] = None,
                 media_type: Optional[str] = None):
        if location:
            super().__init__(location=location, data=data)
        else:
            super().__init__(data=data)

        self.alt = alt
        self._width = width if (width and width > 0) else None
        self._height = height if (height and height > 0) else None
        self._media_type = media_type

    def convert(self, to_type: str = mimetypes.IMAGE_WEBP):
        return Image(data=Processor(self.read()).convert(to_type).get())

    @property
    def md5(self) -> str:
        from hashlib import md5
        return md5(self.read()).hexdigest()

    @property
    def _img(self):
        from PIL import Image
        return Image.open(BytesIO(self.read()))

    @property
    def width(self) -> int:
        if not self._width:
            self._width = self._img.width
        return self._width

    @property
    def height(self) -> int:
        if not self._height:
            self._height = self._img.height
        return self._height

    @property
    def media_type(self) -> str:
        """
        Loads an image and checks its format. Be careful on loading remote images
        """
        if not self._media_type:
            self._media_type = mimetypes.lookup(self._img.format)
        return self._media_type

    def compress(self):
        return Image(data=Processor(self.read()).compress().get())

    def save(self):
        if ((self.is_local() and self.scheme() != "memory") and
                (self.path.is_dir() and self.path.exists()) or (self.path.parent.exists())):
            self.path.write_bytes(data=self.read())

    def copy_to(self, destination: Path, optimize: bool = False, preserve_filename: bool = True) -> Path | None:
        """
        Copy an image to a destiny and optionally optimize by converting JPG and PNG to WEBP.
        """
        img: Optional[bytes] = None
        target: Optional[Path] = None

        if self.is_local() and (destination and destination.is_dir() and destination.exists()):
            if optimize and (self.media_type == mimetypes.IMAGE_JPEG or
                             self.media_type == mimetypes.IMAGE_PNG):
                img = Processor(self.read()).convert().get()
            else:
                img = self.read()

            if preserve_filename:
                target = destination / str(self.filename)
            else:
                target = destination / self.md5

            if target:
                target.write_bytes(img)

                return target

    def base64(self) -> str:
        import base64
        return f'data:image/png;base64,{base64.b64encode(self.read()).decode()}'



import json
from datetime import datetime
from pathlib import Path

from sqlalchemy import text
from sqlalchemy.exc import DatabaseError
from sqlmodel import SQLModel, Field, Session, select
from commons import runtime

class Migration(SQLModel, table=True):
    filename: str = Field(default=None, primary_key=True, nullable=False)
    executed_at: datetime = Field(default=datetime.now(), nullable=False)


class DatabaseMigrationExecutor:
    def __init__(self, session: Session, path: Path | str):
        self.session: Session = session
        self.path: Path = Path(path)

    def run(self) -> list[Migration]:
        pending_migrations_stack: list[Path] = [Path(file) for file in self.path.glob("*.sql")]
        pending_migrations_stack += [Path(file) for file in self.path.glob("*.json")]
        pending_migrations_stack.sort()  # guarantee the execution order of the sql files based in their filenames

        migrations: list[Migration] = [self._execute(migration_file) for migration_file in pending_migrations_stack]
        self.session.close()

        return migrations

    def _execute(self, file: Path) -> Migration | None:
        migration: Migration | None = None

        if not self._get_last_execution_of(file):
            try:
                if file.name.endswith(".json"):
                    self._migrate_json_data_from(file)
                elif file.name.endswith(".sql"):
                    self._migrate_sql_data_from(file)
                else:
                    pass

                migration = Migration(filename=file.name)
                self.session.add(migration)
                self.session.commit()
            except (json.JSONDecodeError, DatabaseError, RuntimeError, Exception) as e:
                self.session.rollback()
                raise RuntimeError(f"Error executing migration {file}: {str(e)}")

    def _get_last_execution_of(self, file: Path) -> Migration | None:
        return self.session.exec(select(Migration).where(Migration.filename == file.name)).first()

    def _migrate_sql_data_from(self, file: Path):
        # Read the file and execute the sql statements
        with file.open(mode="r", encoding="utf-8") as file:
            # Get all statements from the file and execute each one sequentially
            statements: list[str] = [statement.strip() for statement in file.read().split(";")
                                     if isinstance(statement, str) and statement.strip() != '']

            while statements:
                statement: str = statements.pop(0)  # use the list as a queue for performance
                self.session.exec(text(statement))

    def _migrate_json_data_from(self, file: Path) -> Migration | None:
        root: dict = json.loads(file.read_text())
        for cls_signature in root.keys():  # Reads all the classes
            # get the class object to create an instance of it and persist
            module_name, class_name = cls_signature.rsplit('.', 1)
            module = runtime.import_module(module_name)
            cls = getattr(module, class_name)

            for obj in root[cls_signature]:  # Reads the list of class elements
                self.session.add(cls(**obj))



import tempfile
import unittest
from pathlib import Path


class TestSSL(unittest.TestCase):
    def test_local_cert_gen(self):
        from commons.network.http import certs

        files = certs.get_cert(Path(tempfile.mkdtemp()))

        assert files.cert.exists()
        assert files.key.exists()



"""
Superset of built-in `mimetypes` library.
"""

from mimetypes import *
from typing import Optional

# Make all 154 mime types available in this module
for _v in types_map.values():
    globals()[_v.replace("/", "_").upper()] = _v

def lookup(media_type: str) -> Optional[str]:
    """Lookup for a media type"""
    def guess_type_from_map(_media_type: str) -> Optional[str]:
        return types_map[_media_type.lower()] if _media_type.startswith(".") else types_map[f".{_media_type.lower()}"]

    mimetype: Optional[str] = None
    operations: list = [guess_type_from_map, guess_type, guess_file_type]

    while (not mimetype) and operations and (operation := operations.pop()):
        result = operation(media_type)
        mimetype = result[0] if (type(result) in [tuple, list]) else result

    return mimetype



import tempfile
import unittest
from datetime import date, datetime, timedelta
from pathlib import Path
from unittest.mock import patch

from commons.locales import Locale, LocaleConfig


class TestLocales(unittest.TestCase):
    def setUp(self):
        self.temp_dir = Path(tempfile.mkdtemp())

    def tearDown(self):
        import shutil
        shutil.rmtree(self.temp_dir)

    def _create_messages_po(self):
        messages_dir = self.temp_dir / "en" / "LC_MESSAGES"
        messages_dir.mkdir(parents=True, exist_ok=True)
        (messages_dir / "messages.po").write_text('msgid "Hello"\nmsgstr "Hello, World!"')

    # ----- Locale Tests -----
    def test_locale_initialization(self):
        locale = Locale("en")
        self.assertIsInstance(locale, Locale)
        self.assertEqual(locale.language, "en")

    def test_locale_with_translations_directory(self):
        self._create_messages_po()
        locale = Locale("en", self.temp_dir)
        self.assertIsInstance(locale, Locale)
        self.assertTrue(locale._translation is not None)

    def test_locale_gettext(self):
        self._create_messages_po()
        locale = Locale("en", self.temp_dir)
        self.assertEqual(locale.gettext("Hello"), "Hello, World!")

    def test_locale_languages_method(self):
        locale = Locale("en")
        result = locale.languages(filter_by=["pt_BR", "es_ES"])
        self.assertEqual(result, ["Brazilian Portuguese", "European Spanish"])

    def test_locale_format_number(self):
        locale = Locale("pt_BR")
        self.assertEqual(locale.format_number(1234.56), "1.234,56")

    def test_locale_format_currency(self):
        locale = Locale("pt_BR")
        self.assertEqual(locale.format_currency(1234.56, "BRL"), "R$\u00A01.234,56")

    def test_locale_format_date(self):
        locale = Locale("pt_BR")
        test_date = date(2025, 8, 4)
        self.assertEqual(locale.format_date(test_date), "4 de ago. de 2025")
        self.assertEqual(locale.format_date(test_date, format="short"), "04/08/2025")

    def test_locale_format_datetime(self):
        locale = Locale("pt_BR")
        test_datetime = datetime(2025, 8, 4, 12, 43, 22)
        self.assertEqual(locale.format_datetime(test_datetime), "4 de ago. de 2025 12:43:22")
        self.assertEqual(locale.format_datetime(test_datetime, format="short"), "04/08/2025 12:43")

    def test_locale_format_timedelta(self):
        locale = Locale("pt_BR")
        delta = timedelta(days=4)
        self.assertEqual(locale.format_timedelta(delta), "4 dias")

    def test_invalid_locale_raises_error(self):
        with self.assertRaises(ValueError):
            Locale("invalid")

    # ----- LocaleConfig Tests -----
    def test_locale_config_initialization(self):
        config = LocaleConfig(supported_locales=["en"])
        self.assertIsInstance(config, LocaleConfig)

    def test_locale_config_default_locale(self):
        config = LocaleConfig(supported_locales=["en", "en_US"])
        self.assertEqual(config.default_locale.language, "en")

    def test_locale_config_supported_languages(self):
        config = LocaleConfig(supported_locales=["en", "en_US"])
        self.assertEqual(config.supported_languages, ["English", "American English"])

    def test_locale_config_locales(self):
        config = LocaleConfig(supported_locales=["en", "en_US"])
        self.assertEqual(list(config.locales.keys()), ["en", "en_US"])

    def test_locale_config_lookup(self):
        config = LocaleConfig(supported_locales=["en"])
        self.assertEqual(str(config.lookup("en_US")), "en")

        config = LocaleConfig(supported_locales=["en_US"])
        self.assertEqual(str(config.lookup("en")), "en_US")

    def test_locale_config_invalid_supported_locales(self):
        with self.assertRaises(ValueError):
            LocaleConfig(supported_locales=[])

        with self.assertRaises(ValueError):
            LocaleConfig(supported_locales=["invalid_locale"])

    # ----- Edge Cases -----
    def test_locale_without_translations(self):
        locale = Locale("en", Path("/invalid/path"))
        self.assertEqual(locale.gettext("Hello"), "Hello")

    def test_locale_config_with_nonexistent_translations_dir(self):
        config = LocaleConfig(supported_locales=["en"], translations_directory=Path("/invalid/path"))
        self.assertFalse(hasattr(config, "translations_directory"))

    @patch('commons.locales.Locale.compile')
    def test_locale_compile_method(self, mock_compile):
        messages_dir = self.temp_dir / "fr" / "LC_MESSAGES"
        messages_dir.mkdir(parents=True, exist_ok=True)
        (messages_dir / "messages.po").write_text("msgid 'Bonjour'")
        Locale("fr", self.temp_dir)
        mock_compile.assert_called_once_with(messages_dir)

    def test_locale_format_interval(self):
        locale = Locale("pt_BR")
        start = date(2025, 8, 4)
        end = date(2025, 10, 8)
        self.assertEqual(
            locale.format_interval(start, end, skeleton="ddMMyyyy"),
            "04/08/2025\u2009–\u200908/10/2025"
        )

    def test_locale_format_list(self):
        locale = Locale("pt_BR")
        self.assertEqual(locale.format_list(["1", "2", "3"]), "1, 2 e 3")



from datetime import datetime, timedelta

from pydantic import computed_field
from sqlmodel import SQLModel, Field, Session, select

from commons.database import DatabaseAdapter


class CacheEntry(SQLModel, table=True):
    __tablename__ = "cache"
    id: str = Field(nullable=False, primary_key=True)
    data: bytes | None = Field(nullable=True)
    created_at: datetime = Field(default_factory=datetime.now, nullable=False)
    max_age: int = Field(default=300, nullable=False)

    @computed_field
    @property
    def expires(self) -> datetime:
        return self.created_at + timedelta(seconds=self.max_age)

    @computed_field
    @property
    def expired(self) -> bool:
        return bool(datetime.now() >= self.expires)


class Cache:
    def __init__(self, database: DatabaseAdapter):
        self.database = database
        self.session: Session | None = None

    def get(self, key: str) -> CacheEntry | None:
        return self.session.exec(select(CacheEntry).where(CacheEntry.id == key)).first()

    def set(self, key: str, value: bytes, max_age: int = 300) -> CacheEntry:
        entry: CacheEntry = self.get(key)

        if (entry and entry.expired) or (not entry):
            entry = CacheEntry(id=key, data=value, max_age=max_age)
            self.session.add(entry)
            self.session.commit()

        return entry

    def invalidate(self, key: str):
        entry = self.get(key)
        if entry:
            self.session.delete(entry)

    def __enter__(self):
        self.session = self.database.session()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()
        self.session = None



from datetime import datetime, date, timedelta, time, tzinfo
from decimal import Decimal
from pathlib import Path
from typing import Optional, Literal

from babel.core import UnknownLocaleError
from babel import Locale as _BabelLocale


class Locale:
    _locale: Optional[_BabelLocale] = None

    def __init__(self, locale: str, root_directory: Optional[Path] = None):
        try:
            self._locale: _BabelLocale = _BabelLocale.parse(locale)
        except UnknownLocaleError as e:
            raise ValueError(e)

        if root_directory and root_directory.exists():
            import gettext
            self.compile(root_directory / str(self) / "LC_MESSAGES")
            self._translation = gettext.translation("messages", localedir=str(root_directory), languages=[str(self)], fallback=True)
        else:
            self._translation = None

    def __str__(self):
        return str(self._locale)

    @property
    def language(self) -> str:
        return self._locale.language

    @property
    def territory(self) -> Optional[str]:
        return self._locale.territory

    @property
    def display_name(self) -> str:
        return self._locale.display_name

    def languages(self, filter_by: list[str]) -> list[str]:
        return [self._locale.languages.get(lang) for lang in filter_by if lang]

    def format_number(self, number: int | float | Decimal, **kwargs) -> str:
        from babel.numbers import format_decimal
        return format_decimal(number=number, locale=self._locale, **kwargs)

    def format_currency(self, number: int | float | Decimal, currency_code: str, **kwargs) -> str:
        from babel.numbers import format_currency
        return format_currency(number=number, currency=currency_code, locale=self._locale, **kwargs)

    def format_timezone(self, timezone: str | tzinfo) -> str:
        from babel.dates import get_timezone_name
        if isinstance(timezone, str):
            # If timezone is a string, use it directly
            return get_timezone_name(timezone, locale=self._locale)
        elif isinstance(timezone, tzinfo):
            # If timezone is a tzinfo object, get its timezone name
            if hasattr(timezone, "zone"):
                return get_timezone_name(timezone.zone, locale=self._locale)
            else:
                raise ValueError("Unsupported tzinfo object")
        else:
            raise TypeError("Timezone must be a string or a tzinfo object")

    def format_datetime(self, value: datetime, format: Literal['full', 'long', 'medium', 'short'] | str = 'medium',
                        tzinfo: datetime.tzinfo = None) -> str:
        from babel.dates import format_datetime
        return format_datetime(datetime=value, locale=self._locale, format=format, tzinfo=tzinfo)

    def format_date(self, value: date, format: Literal['full', 'long', 'medium', 'short'] | str = 'medium') -> str:
        from babel.dates import format_date
        return format_date(date=value, locale=self._locale, format=format)

    def format_time(self, value: time, format: Literal['full', 'long', 'medium', 'short'] | str = 'medium',
                    tzinfo: datetime.tzinfo = None) -> str:
        from babel.dates import format_time
        return format_time(time=value, locale=self._locale, format=format, tzinfo=tzinfo)

    def format_interval(self, start: date | time, end: date | time, skeleton: Optional[str] = None,
                        tzinfo: datetime.tzinfo = None, **kwargs):
        from babel.dates import format_interval
        return format_interval(start=start, end=end, skeleton=skeleton, locale=self._locale, tzinfo=tzinfo, **kwargs)

    def format_timedelta(self, value: timedelta,
                         granularity: Literal['year', 'month', 'week', 'day', 'hour', 'minute', 'second'] = 'second',
                         format: Literal['narrow', 'short', 'medium', 'long'] = 'long', **kwargs):
        from babel.dates import format_timedelta
        return format_timedelta(delta=value, locale=self._locale, granularity=granularity, format=format, **kwargs)

    def format_list(self, items: list,
                    style: Literal['standard', 'standard-short', 'or', 'or-short', 'unit', 'unit-short', 'unit-narrow'] = 'standard'):
        from babel.lists import format_list
        return format_list(lst=items, locale=self._locale, style=style)

    @classmethod
    def compile(cls, directory: Path):
        # requires gettext at runtime
        if (directory / "messages.po").exists():
            from command_runner import command_runner
            command_runner(f"pybabel compile --directory={directory} --use-fuzzy "
                           f"--input-file={directory / 'messages.po'} "
                           f"--output-file={directory / 'messages.mo'}", live_output=True)

    def gettext(self, msgid: str):
        if self._translation:
            return self._translation.gettext(msgid)
        else:
            return msgid


class LocaleConfig:
    translations_directory: Path
    locales: dict[str, Locale]
    default_locale: Locale

    def __init__(self, supported_locales: list[str], translations_directory: Optional[Path] = None):
        if translations_directory and translations_directory.exists():
            self.translations_directory = translations_directory

        if supported_locales:
            self.locales = {}

            for locale in supported_locales:
                locale = Locale(locale, translations_directory)

                if locale and locale.territory:
                    # guarantee that both language and language + territory are registered as supported locales
                    self.locales[locale.language] = locale

                self.locales[str(locale)] = locale

            self.default_locale = self.locales.get(supported_locales[0])
        else:
            raise ValueError("Supported Locales were not provided.")

    def lookup(self, locale: str | Locale) -> Optional[Locale]:
        item: Optional[Locale] = None

        if type(locale) == str:
            locale = Locale(locale)

        item = self.locales.get(str(locale))
        if not item and locale.territory:
            item = self.locales.get(locale.language)

        return item

    @property
    def supported_languages(self) -> list[str]:
        return self.default_locale.languages(filter_by=[str(locale) for locale in self.locales])

    def extract_translatable_strings(self, directory: Path):
        raise NotImplementedError



import logging
import sys
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Optional, Union


def config(
        level: int = logging.INFO,
        format: str = '%(asctime)s.%(msecs)03d - %(name)s [%(levelname)s]: %(message)s',
        datefmt: Optional[str] = "%Y-%m-%d %H:%M:%S",  # ISO 8601 format,
        directory: Optional[Union[Path, str]] = None,
        max_file_bytes: int = 1024 * 1024,  # 1 MB
        backup_count: int = 5,
        stream=sys.stdout,
) -> logging.Logger:
    """
    Configure global logging settings, including file logging if a directory is provided.

    :param level: The logging level (e.g., logging.INFO, logging.DEBUG).
    :param format: The logging format string.
    :param datefmt: The date/time format string.
    :param app_name: The name of the application. Used for naming the log file.
    :param directory: The directory where log files will be stored.
    :param max_file_bytes: The maximum size of a log file before it is rotated.
    :param backup_count: The number of backup log files to keep.
    :param stream: A stream to use for logging (e.g., sys.stdout, sys.stderr).
    """
    # Get the root logger
    logger = logging.getLogger()

    # Configure handlers for the root logger
    # Remove existing handlers to avoid duplication
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)
        handler.close()

    # Create a formatter
    formatter = logging.Formatter(format, datefmt)

    # Add a console handler if a stream is not provided
    if stream is not None:
        stream_handler = logging.StreamHandler(stream)
        stream_handler.setLevel(level)
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

    # Add a file handler if a directory is provided
    if directory:
        directory = Path(directory)
        if not directory.exists():
            directory.mkdir(exist_ok=True, parents=True)
        log_file = directory / f"{logger.name}.log"
        file_handler = RotatingFileHandler(
            log_file,
            maxBytes=max_file_bytes,
            backupCount=backup_count,
        )
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    logger.setLevel(level)

    return logger


def get_logger(
        name: Optional[str | type] = None,
        level: Optional[int] = None,
        propagate: bool = True
) -> logging.Logger:
    """
    Get a logger with the specified name. Inherits configurations from the global logger by default,
    but can be configured with distinct settings.

    :param name: The name of the logger.
    :param level: The logging level. If None, inherits from the global logger.
    :param propagate: Whether to propagate messages to parent loggers.
    :return: The configured logger instance.
    """
    from inspect import getmodule

    if name is None:
        logger = logging.getLogger()  # get root logger
    else:
        if isinstance(name, type):
            name = f"{getmodule(name).__name__}.{name.__name__}"
        else:
            name = str(name)
        logger = logging.getLogger(name)

    # Set the logging level if provided
    if level is not None:
        logger.setLevel(level)

    # Configure propagation
    logger.propagate = propagate

    return logger



import logging
import tempfile
import unittest
from pathlib import Path

from commons.database import DatabaseAdapter

TEMP_DIR: Path = Path(tempfile.mkdtemp())
logging.getLogger(__file__).warning(f"Temporary directory: {TEMP_DIR.resolve()}")


class TestModels(unittest.TestCase):
    @staticmethod
    def _get_db() -> DatabaseAdapter:
        return DatabaseAdapter(
            scheme="sqlite",
            database=tempfile.mktemp()
        )

    def test_generic_resource(self):
        from commons.models import Resource
        from sqlmodel import SQLModel, Field, select

        def uuid_factory():
            import uuid
            return str(uuid.uuid4())

        class ChildResource(Resource, table=True):
            id: str = Field(primary_key=True, nullable=False, default_factory=uuid_factory)

        db = self._get_db()
        resource = ChildResource()

        assert resource and resource.is_local() and resource.scheme() == "memory"
        assert not resource.read()

        SQLModel.metadata.create_all(db.engine())
        with db.session() as session:
            session.add(resource)
            assert session.exec(select(ChildResource)).first()
            session.commit()
            session.close()

    def test_pix(self):
        from commons.currencies.payments import Pix
        from commons.media import Image

        pix: Pix = Pix(
            receiver_name="Sample Person",
            key="pix@mail.com",
            amount=0.01
        )

        assert pix.brcode()
        image: Image = pix.qrcode()
        assert image
        assert image.base64()

    def test_image(self):
        from PIL import Image as PILImage
        from commons.media import Image
        from commons.media import mimetypes
        from io import BytesIO

        # Create a 1x1 pixel image with a transparent background
        png_img = PILImage.new("RGBA", (1, 1), (0, 0, 0, 0))  # RGBA mode for transparency
        buffer = BytesIO()
        img: Image = Image()

        png_img.save(buffer, "png")
        img.data=buffer.getvalue()

        assert img.media_type == mimetypes.IMAGE_PNG
        assert img.width == 1
        assert img.height == 1
        assert img.is_local()
        assert img.base64()

        _d = img.model_dump(exclude={"url", "path"})
        _d["location"] = tempfile.mktemp()
        img2 = Image(**_d)
        img2.save()
        assert img2.path.exists()
